/*
 * main.c -- a random Go player
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "common.h"
#include "log.h"
#include "memory.h"
#include "random.h"


/**************************************************************/


#define LOG_FILE	"randomGo.log"

#define LINE_SIZE	1000
#define MAX_TOKENS	50

#define MIN_BOARD	5
#define MAX_BOARD	21

#define BIG_BOARD(n)	((n) > 19)
#define SIZE19_BOARD(n)	((n) == 19)
#define MEDIUM_BOARD(n)	((n) >= 13 && (n) <= 18)
#define SMALL_BOARD(n)	((n) >= 8 && (n) <= 12)
#define SIZE7_BOARD(n)	((n) == 7)
#define TINY_BOARD(n)	((n) < 7)

#define EVEN_BOARD(n)	(((n) & 1) == 0)

#define ON_BOARD(r,c)	((r) >= 0 && (r) < boardSize && \
			 (c) >= 0 && (c) < boardSize)


/**************************************************************/


typedef int Color;

#define NONE		0
#define WHITE		1
#define BLACK		2
#define BOTH		3

#define OPPONENT(c)	((c) == WHITE ? BLACK : WHITE)


typedef struct {
  Color grid[MAX_BOARD][MAX_BOARD];
  int capturedWhite;
  int capturedBlack;
} Board;


typedef struct {
  Color color;
  int row;
  int col;
} Move;


#define MOVE_SET	-1
#define MOVE_PASS	-2
#define MOVE_RESIGN	-3

#define IS_SET(m)	((m).row == MOVE_SET)
#define IS_PASS(m)	((m).row == MOVE_PASS)
#define IS_RESIGN(m)	((m).row == MOVE_RESIGN)

#define MK_SET(m)	((m).row = MOVE_SET)
#define MK_PASS(m)	((m).row = MOVE_PASS)
#define MK_RESIGN(m)	((m).row = MOVE_RESIGN)


/**************************************************************/


int boardSize;

float komi;

int mainTime;
int byoTime;
int byoStones;
int mainTimeWhite;
int byoTimeWhite;
int byoStonesWhite;
int mainTimeBlack;
int byoTimeBlack;
int byoStonesBlack;

Bool noSuicide;
Bool superKo;

Board board;


/**************************************************************/


void resetClocks(void) {
  mainTimeWhite = mainTime;
  byoTimeWhite = byoTime;
  byoStonesWhite = 0;
  mainTimeBlack = mainTime;
  byoTimeBlack = byoTime;
  byoStonesBlack = 0;
}


/**************************************************************/


int mapRownumToInt(int row) {
  return row + 1;
}


int mapIntToRownum(int n) {
  return n - 1;
}


char mapColnumToChar(int col) {
  char result;

  result = 'A' + col;
  if (result >= 'I') {
    result++;
  }
  return result;
}


int mapCharToColnum(char c) {
  int result;

  if (c >= 'a' && c <= 'z') {
    c += 'A' - 'a';
  }
  result = c - 'A';
  if (result >= 8) {
    result--;
  }
  return result;
}


char *placeFixedHandicap(Board *board, int handicap) {
  static char buffer[LINE_SIZE];
  int lo, md, hi;
  char *p;

  if (TINY_BOARD(boardSize)) {
    /* no handicap stones for tiny boards */
    return NULL;
  }
  if (SIZE7_BOARD(boardSize) || SMALL_BOARD(boardSize)) {
    /* handicap stones placed on third line */
    lo = 2;
    md = (boardSize - 1) / 2;
    hi = boardSize - 3;
  } else {
    /* handicap stones placed on fourth line */
    lo = 3;
    md = (boardSize - 1) / 2;
    hi = boardSize - 4;
  }
  p = buffer;
  if (handicap >= 2) {
    board->grid[lo][lo] = BLACK;
    board->grid[hi][hi] = BLACK;
    p += sprintf(p, "%c%d %c%d",
                 mapColnumToChar(lo), mapRownumToInt(lo),
                 mapColnumToChar(hi), mapRownumToInt(hi));
  }
  if (handicap >= 3) {
    board->grid[hi][lo] = BLACK;
    p += sprintf(p, " %c%d",
                 mapColnumToChar(lo), mapRownumToInt(hi));
  }
  if (handicap >= 4) {
    board->grid[lo][hi] = BLACK;
    p += sprintf(p, " %c%d",
                 mapColnumToChar(hi), mapRownumToInt(lo));
  }
  if (handicap >= 6) {
    board->grid[md][lo] = BLACK;
    board->grid[md][hi] = BLACK;
    p += sprintf(p, " %c%d %c%d",
                 mapColnumToChar(lo), mapRownumToInt(md),
                 mapColnumToChar(hi), mapRownumToInt(md));
  }
  if (handicap >= 8) {
    board->grid[lo][md] = BLACK;
    board->grid[hi][md] = BLACK;
    p += sprintf(p, " %c%d %c%d",
                 mapColnumToChar(md), mapRownumToInt(lo),
                 mapColnumToChar(md), mapRownumToInt(hi));
  }
  if (handicap == 5 ||
      handicap == 7 ||
      handicap == 9) {
    board->grid[md][md] = BLACK;
    p += sprintf(p, " %c%d",
                 mapColnumToChar(md), mapRownumToInt(md));
  }
  p += sprintf(p, "\n");
  return buffer;
}


Bool isStarPoint(int row, int col) {
  int lo, md, hi;

  if (TINY_BOARD(boardSize)) {
    /* no star points on tiny boards */
    return FALSE;
  }
  if (SIZE7_BOARD(boardSize)) {
    /* 4 star points on a 7x7 board */
    lo = 2;
    hi = boardSize - 3;
    return (row == lo && col == lo) ||
           (row == lo && col == hi) ||
           (row == hi && col == lo) ||
           (row == hi && col == hi);
  }
  if (SMALL_BOARD(boardSize)) {
    /* star points placed on third line */
    lo = 2;
    md = (boardSize - 1) / 2;
    hi = boardSize - 3;
  } else {
    /* star points placed on fourth line */
    lo = 3;
    md = (boardSize - 1) / 2;
    hi = boardSize - 4;
  }
  if (EVEN_BOARD(boardSize)) {
    return (row == lo && col == lo) ||
           (row == lo && col == hi) ||
           (row == hi && col == lo) ||
           (row == hi && col == hi);
  } else {
    return (row == lo && col == lo) ||
           (row == lo && col == md) ||
           (row == lo && col == hi) ||
           (row == md && col == lo) ||
           (row == md && col == md) ||
           (row == md && col == hi) ||
           (row == hi && col == lo) ||
           (row == hi && col == md) ||
           (row == hi && col == hi);
  }
}


char *boardToString(Board *board) {
  static char colorChar[] = { '-', 'O', '#', '+' };
  static char buffer[2 * MAX_BOARD * LINE_SIZE];
  char *p;
  int i, j;
  Color color;

  p = buffer;
  p += sprintf(p, "\n");
  if (boardSize < 10) {
    p += sprintf(p, "  ");
  } else {
    p += sprintf(p, "   ");
  }
  for (j = 0; j < boardSize; j++) {
    p += sprintf(p, "%c ", mapColnumToChar(j));
  }
  p += sprintf(p, "\n");
  for (i = boardSize - 1; i > -1; i--) {
    if (boardSize < 10) {
      p += sprintf(p, "%1d ", mapRownumToInt(i));
    } else {
      p += sprintf(p, "%2d ", mapRownumToInt(i));
    }
    for (j = 0; j < boardSize; j++) {
      color = board->grid[i][j];
      if (color == NONE && isStarPoint(i, j)) {
        color = BOTH;
      }
      p += sprintf(p, "%c ", colorChar[color]);
    }
    if (boardSize < 10) {
      p += sprintf(p, "%1d ", mapRownumToInt(i));
    } else {
      p += sprintf(p, "%2d ", mapRownumToInt(i));
    }
    if (i == (boardSize - 1) / 2 + 2) {
      p += sprintf(p, "  White has captured %d stones",
                   111);
    } else
    if (i == (boardSize - 1) / 2 + 1) {
      p += sprintf(p, "    Time  %3d:%02d / %02d:%02d (%d)",
                   mainTimeWhite / 60, mainTimeWhite % 60,
                   byoTimeWhite / 60, byoTimeWhite % 60,
                   byoStonesWhite);
    } else
    if (i == (boardSize - 1) / 2 - 1) {
      p += sprintf(p, "  Black has captured %d stones",
                   222);
    } else
    if (i == (boardSize - 1) / 2 - 2) {
      p += sprintf(p, "    Time  %3d:%02d / %02d:%02d (%d)",
                   mainTimeBlack / 60, mainTimeBlack % 60,
                   byoTimeBlack / 60, byoTimeBlack % 60,
                   byoStonesBlack);
    }
    p += sprintf(p, "\n");
  }
  if (boardSize < 10) {
    p += sprintf(p, "  ");
  } else {
    p += sprintf(p, "   ");
  }
  for (j = 0; j < boardSize; j++) {
    p += sprintf(p, "%c ", mapColnumToChar(j));
  }
  p += sprintf(p, "\n");
  return buffer;
}


Bool compareBoard(Board *dst, Board *src) {
  int i, j;

  for (i = 0; i < boardSize; i++) {
    for (j = 0; j < boardSize; j++) {
      if (dst->grid[i][j] != src->grid[i][j]) {
        return FALSE;
      }
    }
  }
  return TRUE;
}


void copyBoard(Board *dst, Board *src) {
  int i, j;

  for (i = 0; i < boardSize; i++) {
    for (j = 0; j < boardSize; j++) {
      dst->grid[i][j] = src->grid[i][j];
    }
  }
  dst->capturedWhite = src->capturedWhite;
  dst->capturedBlack = src->capturedBlack;
}


void clearBoard(Board *board) {
  int i, j;

  for (i = 0; i < boardSize; i++) {
    for (j = 0; j < boardSize; j++) {
      board->grid[i][j] = NONE;
    }
  }
  board->capturedWhite = 0;
  board->capturedBlack = 0;
}


Bool isEmptyBoard(Board *board) {
  int i, j;

  for (i = 0; i < boardSize; i++) {
    for (j = 0; j < boardSize; j++) {
      if (board->grid[i][j] != NONE) {
        return FALSE;
      }
    }
  }
  return TRUE;
}


/**************************************************************/


typedef struct historyEntry {
  /* node information */
  Move move;				/* move that led to this node */
  Board board;				/* board position after the move */
  /* node administration */
  struct historyEntry *nextSibling;	/* forward chain of siblings */
  struct historyEntry *prevSibling;	/* backward chain of siblings */
  struct historyEntry *firstChild;	/* first of any children */
  struct historyEntry *lastChild;	/* last of any children */
  struct historyEntry *parent;		/* back to parent node */
} HistoryEntry;


static HistoryEntry *history;
static HistoryEntry *current;


void resetHistory(Board *board) {
  HistoryEntry *entry;

  entry = memoryAllocate(sizeof(HistoryEntry), MEMORY_STATIC);
  entry->move.color = (isEmptyBoard(board) ? WHITE : BLACK);
  entry->move.row = MOVE_SET;
  copyBoard(&entry->board, board);
  entry->nextSibling = NULL;
  entry->prevSibling = NULL;
  entry->firstChild = NULL;
  entry->lastChild = NULL;
  entry->parent = NULL;
  history = entry;
  current = entry;
}


void addToHistory(Move move, Board *board) {
  HistoryEntry *entry;

  entry = memoryAllocate(sizeof(HistoryEntry), MEMORY_STATIC);
  entry->move = move;
  copyBoard(&entry->board, board);
  if (current->firstChild == NULL) {
    current->firstChild = entry;
    entry->nextSibling = NULL;
    entry->prevSibling = NULL;
    current->lastChild = entry;
  } else {
    current->lastChild->nextSibling = entry;
    entry->nextSibling = NULL;
    entry->prevSibling = current->lastChild;
    current->lastChild = entry;
  }
  entry->firstChild = NULL;
  entry->lastChild = NULL;
  entry->parent = current;
  current = entry;
}


void getCurrentFromHistory(Board *board) {
  copyBoard(board, &current->board);
}


Bool goBackInHistory(void) {
  if (current->parent == NULL) {
    return FALSE;
  }
  current = current->parent;
  return TRUE;
}


Bool simpleKoViolation(Board *board) {
  if (current->parent == NULL) {
    return FALSE;
  }
  return compareBoard(&current->parent->board, board);
}


Bool superKoViolation(Board *board) {
  HistoryEntry *entry;

  entry = current->parent;
  while (entry != NULL) {
    if (compareBoard(&entry->board, board)) {
      return TRUE;
    }
    entry = entry->parent;
  }
  return FALSE;
}


/**************************************************************/


int countRecursive(Color grid[MAX_BOARD][MAX_BOARD],
                   Bool flags[MAX_BOARD][MAX_BOARD],
                   Color color, int row, int col) {
  if (!ON_BOARD(row, col)) {
    /* off board doesn't count */
    return 0;
  }
  if (flags[row][col]) {
    /* already counted */
    return 0;
  }
  if (grid[row][col] == NONE) {
    /* this is a liberty */
    flags[row][col] = TRUE;
    return 1;
  }
  if (grid[row][col] == color) {
    /* a stone of our color */
    flags[row][col] = TRUE;
    return countRecursive(grid, flags, color, row - 1, col) +
           countRecursive(grid, flags, color, row + 1, col) +
           countRecursive(grid, flags, color, row, col - 1) +
           countRecursive(grid, flags, color, row, col + 1);
  }
  /* a stone of our opponent */
  return 0;
}


int countLiberties(Board *board, int row, int col) {
  int liberties;
  Bool flags[MAX_BOARD][MAX_BOARD];
  int i, j;

  for (i = 0; i < boardSize; i++) {
    for (j = 0; j < boardSize; j++) {
      flags[i][j] = FALSE;
    }
  }
  liberties = countRecursive(board->grid, flags,
                             board->grid[row][col], row, col);
  logMessage("liberties for string at %c%d: %d\n",
             mapColnumToChar(col), mapRownumToInt(row), liberties);
  return liberties;
}


int removeRecursive(Color grid[MAX_BOARD][MAX_BOARD],
                    Color color, int row, int col) {
  int prisoners;

  if (!ON_BOARD(row, col)) {
    return 0;
  }
  prisoners = 0;
  if (grid[row][col] == color) {
    grid[row][col] = NONE;
    prisoners++;
    prisoners += removeRecursive(grid, color, row - 1, col);
    prisoners += removeRecursive(grid, color, row + 1, col);
    prisoners += removeRecursive(grid, color, row, col - 1);
    prisoners += removeRecursive(grid, color, row, col + 1);
  }
  return prisoners;
}


void removeString(Board *board, int row, int col) {
  removeRecursive(board->grid, board->grid[row][col], row, col);
}


/**************************************************************/


Bool checkMove(Board *board, Move move) {
  Color color;
  int row, col;
  int opponent;

  /* if legal move: board holds new position */
  /* if illegal move: board may be clobbered */
  if (IS_PASS(move)) {
    return TRUE;
  }
  color = move.color;
  row = move.row;
  col = move.col;
  if (!ON_BOARD(row, col)) {
    return FALSE;
  }
  if (board->grid[row][col] != NONE) {
    return FALSE;
  }
  board->grid[row][col] = color;
  opponent = OPPONENT(color);
  if (ON_BOARD(row - 1, col) &&
      board->grid[row - 1][col] == opponent &&
      countLiberties(board, row - 1, col) == 0) {
    removeString(board, row - 1, col);
  }
  if (ON_BOARD(row + 1, col) &&
      board->grid[row + 1][col] == opponent &&
      countLiberties(board, row + 1, col) == 0) {
    removeString(board, row + 1, col);
  }
  if (ON_BOARD(row, col - 1) &&
      board->grid[row][col - 1] == opponent &&
      countLiberties(board, row, col - 1) == 0) {
    removeString(board, row, col - 1);
  }
  if (ON_BOARD(row, col + 1) &&
      board->grid[row][col + 1] == opponent &&
      countLiberties(board, row, col + 1) == 0) {
    removeString(board, row, col + 1);
  }
  if (countLiberties(board, row, col) == 0) {
    /* this move is suicide */
    if (noSuicide) {
      return FALSE;
    }
    removeString(board, row, col);
  }
  if (superKo) {
    return !superKoViolation(board);
  } else {
    return !simpleKoViolation(board);
  }
}


Bool playMove(Move m) {
  if (checkMove(&board, m)) {
    addToHistory(m, &board);
    return TRUE;
  } else {
    getCurrentFromHistory(&board);
    return FALSE;
  }
}


/**************************************************************/


Move genMove(Color color) {
  Move m;
  int n;

  m.color = color;
  for (n = 0; n < 100; n++) {
    m.row = randomNext() % boardSize;
    m.col = randomNext() % boardSize;
    if (playMove(m)) {
      return m;
    }
  }
  /* no legal move found */
  MK_PASS(m);
  return m;
}


/**************************************************************/


Bool run;
Bool idPresent;
int id;


typedef struct {
  char *name;
  int minArgs;
  void (*func)(int argc, char *argv[]);
} Cmd;

extern Cmd cmds[];
extern int numCmds;


Cmd *lookupCmd(char *name);


void reply(Bool succ, char *answer) {
  char idBuffer[20];

  if (idPresent) {
    sprintf(idBuffer, "%d", id);
  } else {
    idBuffer[0] = '\0';
  }
  logMessage("GTP reply: %s%s %s\n",
             succ ? "=" : "?", idBuffer,
             (answer == NULL || *answer == '\0') ? "\n" : answer);
  printf("%s%s %s\n",
         succ ? "=" : "?", idBuffer,
         (answer == NULL || *answer == '\0') ? "\n" : answer);
  fflush(stdout);
}


void syntaxError(void) {
  reply(FALSE, "syntax error\n");
}


int strcmpIgnore(char *s1, char *s2) {
  while (tolower(*s1) == tolower(*s2)) {
    if (*s1 == '\0') {
      return 0;
    }
    s1++;
    s2++;
  }
  return (unsigned char) tolower(*s1) -
         (unsigned char) tolower(*s2);
}


Bool parseInt(char *str, int *i) {
  char *endptr;

  *i = strtoul(str, &endptr, 10);
  return *endptr == '\0';
}


Bool parseFloat(char *str, float *f) {
  char *endptr;

  *f = strtod(str, &endptr);
  return *endptr == '\0';
}


Bool parseColor(char *str, Color *color) {
  if (strcmpIgnore(str, "white") == 0) {
    *color = WHITE;
    return TRUE;
  }
  if (strcmpIgnore(str, "w") == 0) {
    *color = WHITE;
    return TRUE;
  }
  if (strcmpIgnore(str, "black") == 0) {
    *color = BLACK;
    return TRUE;
  }
  if (strcmpIgnore(str, "b") == 0) {
    *color = BLACK;
    return TRUE;
  }
  return FALSE;
}


Bool parseVertex(char *str, Move *m) {
  int n;

  if (strcmpIgnore(str, "pass") == 0) {
    MK_PASS(*m);
    return TRUE;
  }
  if ((str[0] >= 'A' && str[0] <= 'Z') ||
      (str[0] >= 'a' && str[0] <= 'z')) {
    m->col = mapCharToColnum(str[0]);
  } else {
    return FALSE;
  }
  if (parseInt(&str[1], &n)) {
    m->row = mapIntToRownum(n);
  } else {
    return FALSE;
  }
  return TRUE;
}


void do_protocol_version(int argc, char *argv[]) {
  reply(TRUE, "2\n");
}


void do_name(int argc, char *argv[]) {
  reply(TRUE, "randomGo\n");
}


void do_version(int argc, char *argv[]) {
  reply(TRUE, "0.0\n");
}


void do_known_command(int argc, char *argv[]) {
  Cmd *cmd;

  cmd = lookupCmd(argv[0]);
  if (cmd == NULL) {
    reply(TRUE, "false\n");
  } else {
    reply(TRUE, "true\n");
  }
}


void do_list_commands(int argc, char *argv[]) {
  char buffer[50 * LINE_SIZE];
  char *p;
  int i;

  p = buffer;
  for (i = 0; i < numCmds; i++) {
    p += sprintf(p, "%s\n", cmds[i].name);
  }
  reply(TRUE, buffer);
}


void do_quit(int argc, char *argv[]) {
  run = FALSE;
  reply(TRUE, NULL);
}


void do_boardsize(int argc, char *argv[]) {
  int size;

  if (!parseInt(argv[0], &size)) {
    syntaxError();
  } else
  if (size < MIN_BOARD || size > MAX_BOARD) {
    reply(FALSE, "unacceptable size\n");
  } else {
    boardSize = size;
    reply(TRUE, NULL);
  }
}


void do_clear_board(int argc, char *argv[]) {
  clearBoard(&board);
  resetHistory(&board);
  reply(TRUE, NULL);
}


void do_komi(int argc, char *argv[]) {
  float value;

  if (!parseFloat(argv[0], &value)) {
    syntaxError();
  } else {
    komi = value;
    reply(TRUE, NULL);
  }
}


void do_fixed_handicap(int argc, char *argv[]) {
  int n;
  char *answer;

  if (!parseInt(argv[0], &n)) {
    syntaxError();
  } else
  if (!isEmptyBoard(&board)) {
    reply(FALSE, "board not empty\n");
  } else
  if (n < 2 || n > 9 ||
      (EVEN_BOARD(boardSize) && n > 4) ||
      (SIZE7_BOARD(boardSize) && n > 4) ||
      TINY_BOARD(boardSize)) {
    reply(FALSE, "invalid number of stones\n");
  } else {
    answer = placeFixedHandicap(&board, n);
    reply(TRUE, answer);
  }
}


void do_place_free_handicap(int argc, char *argv[]) {
}


void do_set_free_handicap(int argc, char *argv[]) {
}


void do_play(int argc, char *argv[]) {
  Move m;

  if (!parseColor(argv[0], &m.color) ||
      !parseVertex(argv[1], &m)) {
    syntaxError();
  } else
  if (!playMove(m)) {
    reply(FALSE, "illegal move\n");
  } else {
    reply(TRUE, NULL);
  }
}


void do_genmove(int argc, char *argv[]) {
  Color color;
  Move move;
  char answer[LINE_SIZE];

  if (!parseColor(argv[0], &color)) {
    syntaxError();
  } else {
    move = genMove(color);
    if (IS_RESIGN(move)) {
      sprintf(answer, "resign\n");
    } else
    if (IS_PASS(move)) {
      sprintf(answer, "pass\n");
    } else {
      sprintf(answer, "%c%d\n",
              mapColnumToChar(move.col),
              mapRownumToInt(move.row));
    }
    reply(TRUE, answer);
  }
}


void do_undo(int argc, char *argv[]) {
}


void do_time_settings(int argc, char *argv[]) {
  int mTime;
  int bTime;
  int bStones;

  if (!parseInt(argv[0], &mTime) ||
      !parseInt(argv[1], &bTime) ||
      !parseInt(argv[2], &bStones)) {
    syntaxError();
  } else {
    mainTime = mTime;
    byoTime = bTime;
    byoStones = bStones;
    resetClocks();
    reply(TRUE, NULL);
  }
}


void do_time_left(int argc, char *argv[]) {
  Color color;
  int time;
  int stones;

  if (!parseColor(argv[0], &color) ||
      !parseInt(argv[1], &time) ||
      !parseInt(argv[2], &stones)) {
    syntaxError();
  } else {
    if (color == WHITE) {
      /* inform white about time */
      if (stones == 0) {
        /* main time */
        mainTimeWhite = time;
      } else {
        /* byo-yomi time and stones */
        byoTimeWhite = time;
        byoStonesWhite = stones;
      }
    } else {
      /* inform black about time */
      if (stones == 0) {
        /* main time */
        mainTimeBlack = time;
      } else {
        /* byo-yomi time and stones */
        byoTimeBlack = time;
        byoStonesBlack = stones;
      }
    }
    reply(TRUE, NULL);
  }
}


void do_final_score(int argc, char *argv[]) {
}


void do_final_status_list(int argc, char *argv[]) {
}


void do_showboard(int argc, char *argv[]) {
  reply(TRUE, boardToString(&board));
}


Cmd cmds[] = {
  { "protocol_version",    0, do_protocol_version    },
  { "name",                0, do_name                },
  { "version",             0, do_version             },
  { "known_command",       1, do_known_command       },
  { "list_commands",       0, do_list_commands       },
  { "quit",                0, do_quit                },
  { "boardsize",           1, do_boardsize           },
  { "clear_board",         0, do_clear_board         },
  { "komi",                1, do_komi                },
  { "fixed_handicap",      1, do_fixed_handicap      },
  //{ "place_free_handicap", 9, do_place_free_handicap },
  //{ "set_free_handicap",   9, do_set_free_handicap   },
  { "play",                2, do_play                },
  { "genmove",             1, do_genmove             },
  //{ "undo",                9, do_undo                },
  { "time_settings",       3, do_time_settings       },
  { "time_left",           3, do_time_left           },
  //{ "final_score",         9, do_final_score         },
  //{ "final_status_list",   9, do_final_status_list   },
  { "showboard",           0, do_showboard           }
};


int numCmds = sizeof(cmds)/sizeof(cmds[0]);


/**************************************************************/


void preprocessLine(char *line) {
  char *p;
  char *q;
  char c;

  p = line;
  q = line;
  while (1) {
    c = *p++;
    if (c == '\0' || c == '#') {
      *q = '\0';
      break;
    }
    if (c < 32 || c == 127) {
      if (c == '\t') {
        *q++ = ' ';
      }
      continue;
    }
    *q++ = c;
  }
}


int tokenizeLine(char *line, char *tokens[], int maxTokens) {
  int n;
  char *p;

  n = 0;
  p = strtok(line, " ");
  while (p != NULL) {
    if (n < MAX_TOKENS) {
      tokens[n++] = p;
    }
    p = strtok(NULL, " ");
  }
  return n;
}


Cmd *lookupCmd(char *name) {
  Cmd *p;
  int i;

  p = cmds;
  for (i = 0; i < numCmds; i++) {
    if (strcmp(p->name, name) == 0) {
      /* cmd found */
      return p;
    }
    p++;
  }
  /* not found */
  return NULL;
}


/**************************************************************/


void resetAll(void) {
  boardSize = 19;
  clearBoard(&board);
  resetHistory(&board);
  komi = 6.5;
  mainTime = 25 * 60;
  byoTime = 10 * 60;
  byoStones = 25;
  resetClocks();
  noSuicide = TRUE;
  superKo = FALSE;
}


int main(void) {
  char line[LINE_SIZE];
  char *tokens[MAX_TOKENS];
  int numTokens;
  Cmd *cmd;
  int numArgs;
  char **argVect;

  logOpen(LOG_FILE);
  randomSet(0);
  memoryInit(LOG1M, MEMORY_CLUSTERS);
  resetAll();
  run = TRUE;
  while (run) {
    /* read command */
    if (fgets(line, LINE_SIZE, stdin) == NULL) {
      break;
    }
    logMessage("GTP request: %s", line);
    /* do preprocessing and tokenizing */
    preprocessLine(line);
    numTokens = tokenizeLine(line, tokens, MAX_TOKENS);
    if (numTokens == 0) {
      /* no tokens at all: ignore line */
      continue;
    }
    if (numTokens == MAX_TOKENS) {
      /* we may have discarded tokens already */
      syntaxError();
      continue;
    }
    /* handle optional cmd id */
    idPresent = parseInt(tokens[0], &id);
    if (idPresent && numTokens == 1) {
      /* no tokens at all (except id): ignore line */
      continue;
    }
    /* lookup command in table */
    cmd = lookupCmd(tokens[idPresent ? 1 : 0]);
    if (cmd == NULL) {
      /* command not found */
      reply(FALSE, "unknown command\n");
      continue;
    }
    /* command found: execute it */
    numArgs = (idPresent ? numTokens - 2 : numTokens - 1);
    argVect = (idPresent ? tokens + 2 : tokens + 1);
    if (numArgs < cmd->minArgs) {
      /* too few arguments */
      syntaxError();
      continue;
    }
    (cmd->func)(numArgs, argVect);
  }
  logClose();
  return 0;
}
